@@ -23,8 +23,8 @@
 
 #include <boost/container/detail/config_begin.hpp>
 #include <boost/container/container_fwd.hpp>
+#include <boost/container/detail/workaround.hpp>
 #include <boost/intrusive/pack_options.hpp>
-#include <boost/static_assert.hpp>
 
 namespace boost {
 namespace container {
@@ -52,8 +52,8 @@ enum tree_type_enum
 template<tree_type_enum TreeType, bool OptimizeSize>
 struct tree_opt
 {
-   static const boost::container::tree_type_enum tree_type = TreeType;
-   static const bool optimize_size = OptimizeSize;
+   BOOST_STATIC_CONSTEXPR boost::container::tree_type_enum tree_type = TreeType;
+   BOOST_STATIC_CONSTEXPR bool optimize_size = OptimizeSize;
 };
 
 typedef tree_opt<red_black_tree, true> tree_assoc_defaults;
@@ -113,20 +113,33 @@ using tree_assoc_options_t = typename boost::container::tree_assoc_options<Optio
 
 #if !defined(BOOST_CONTAINER_DOXYGEN_INVOKED)
 
-template<bool StoreHash>
+template<bool StoreHash, bool CacheBegin, bool LinearBuckets, bool FastmodBuckets>
 struct hash_opt
 {
-   static const bool store_hash = StoreHash;
+   BOOST_STATIC_CONSTEXPR bool store_hash  = StoreHash;
+   BOOST_STATIC_CONSTEXPR bool cache_begin = CacheBegin;
+   BOOST_STATIC_CONSTEXPR bool linear_buckets = LinearBuckets;
+   BOOST_STATIC_CONSTEXPR bool fastmod_buckets = FastmodBuckets;
 };
 
-typedef hash_opt<false> hash_assoc_defaults;
+typedef hash_opt<false, false, false, false> hash_assoc_defaults;
 
 #endif   //!defined(BOOST_CONTAINER_DOXYGEN_INVOKED)
 
-//!This option setter specifies if node size is optimized
-//!storing rebalancing data masked into pointers for ordered associative containers
+//!This option setter specifies if nodes also store the hash value
+//!so that search and rehashing for hash-expensive types is improved.
+//!This option might degrade performance for easy to hash types (like integers)
 BOOST_INTRUSIVE_OPTION_CONSTANT(store_hash, bool, Enabled, store_hash)
 
+//!This option setter specifies if the container will cache the first
+//!non-empty bucket so that begin() is O(1) instead of searching for the
+//!first non-empty bucket (which can be O(bucket_size()))
+BOOST_INTRUSIVE_OPTION_CONSTANT(cache_begin, bool, Enabled, cache_begin)
+
+BOOST_INTRUSIVE_OPTION_CONSTANT(linear_buckets, bool, Enabled, linear_buckets)
+
+BOOST_INTRUSIVE_OPTION_CONSTANT(fastmod_buckets, bool, Enabled, fastmod_buckets)
+
 //! Helper metafunction to combine options into a single type to be used
 //! by \c boost::container::hash_set, \c boost::container::hash_multiset
 //! \c boost::container::hash_map and \c boost::container::hash_multimap.
@@ -147,7 +160,11 @@ struct hash_assoc_options
       Options...
       #endif
       >::type packed_options;
-   typedef hash_opt<packed_options::store_hash> implementation_defined;
+   typedef hash_opt<packed_options::store_hash
+                   ,packed_options::cache_begin
+                   ,packed_options::linear_buckets
+                   ,packed_options::fastmod_buckets
+                   > implementation_defined;
    /// @endcond
    typedef implementation_defined type;
 };
@@ -183,6 +200,20 @@ struct default_if_void<void, Default>
    typedef Default type;
 };
 
+template<std::size_t N, std::size_t DefaultN>
+struct default_if_zero
+{
+   BOOST_STATIC_CONSTEXPR std::size_t value = N;
+};
+
+template<std::size_t DefaultN>
+struct default_if_zero<0u, DefaultN>
+{
+   BOOST_STATIC_CONSTEXPR std::size_t value = DefaultN;
+};
+
+
+
 #endif
 
 #if !defined(BOOST_CONTAINER_DOXYGEN_INVOKED)
@@ -215,24 +246,17 @@ class default_next_capacity;
 
 typedef vector_opt<void, void> vector_null_opt;
 
-template<class GrowthType, class StoredSizeType>
-struct devector_opt
-   : vector_opt<GrowthType, StoredSizeType>
-{};
-
-typedef devector_opt<void, void> devector_null_opt;
-
 #else    //!defined(BOOST_CONTAINER_DOXYGEN_INVOKED)
 
-//!This growth factor argument specifies that the container should increase it's
+//!This growth factor argument specifies that the container should increase its
 //!capacity a 50% when existing capacity is exhausted.
 struct growth_factor_50{};
 
-//!This growth factor argument specifies that the container should increase it's
+//!This growth factor argument specifies that the container should increase its
 //!capacity a 60% when existing capacity is exhausted.
 struct growth_factor_60{};
 
-//!This growth factor argument specifies that the container should increase it's
+//!This growth factor argument specifies that the container should increase its
 //!capacity a 100% (doubling its capacity) when existing capacity is exhausted.
 struct growth_factor_100{};
 
@@ -340,7 +364,7 @@ template<class GrowthType, std::size_t InplaceAlignment>
 struct small_vector_opt
 {
    typedef GrowthType      growth_factor_type;
-   static const std::size_t inplace_alignment = InplaceAlignment;
+   BOOST_STATIC_CONSTEXPR std::size_t inplace_alignment = InplaceAlignment;
 };
 
 typedef small_vector_opt<void, 0u> small_vector_null_opt;
@@ -406,8 +430,8 @@ BOOST_INTRUSIVE_OPTION_CONSTANT(throw_on_overflow, bool, ThrowOnOverflow, throw_
 template<bool ThrowOnOverflow, std::size_t InplaceAlignment>
 struct static_vector_opt
 {
-   static const bool throw_on_overflow = ThrowOnOverflow;
-   static const std::size_t inplace_alignment = InplaceAlignment;
+   BOOST_STATIC_CONSTEXPR bool throw_on_overflow = ThrowOnOverflow;
+   BOOST_STATIC_CONSTEXPR std::size_t inplace_alignment = InplaceAlignment;
 };
 
 typedef static_vector_opt<true, 0u> static_vector_null_opt;
@@ -449,9 +473,92 @@ using static_vector_options_t = typename boost::container::static_vector_options
 
 #endif
 
+
+////////////////////////////////////////////////////////////////
+//
+//
+//          OPTIONS FOR DEVECTOR CONTAINER
+//
+//
+////////////////////////////////////////////////////////////////
+
+//!Thse options specify the relocation strategy of devector.
+//!
+//!Predefined relocation limits that can be passed as arguments to this option are:
+//!\c boost::container::relocate_on_66
+//!\c boost::container::relocate_on_75
+//!\c boost::container::relocate_on_80
+//!\c boost::container::relocate_on_85
+//!\c boost::container::relocate_on_90
+//!
+//!If this option is not specified, a default will be used by the container.
+//!
+//!Note: Repeated insertions at only one end (only back insertions or only front insertions) usually will
+//!lead to a single relocation when `relocate_on_66` is used and two relocations when `relocate_on_90`
+//!is used.
+
+#if !defined(BOOST_CONTAINER_DOXYGEN_INVOKED)
+
+BOOST_INTRUSIVE_OPTION_CONSTANT(relocate_on, std::size_t, Fraction, free_fraction)
+
+struct relocate_on_66 : public relocate_on<3U>{};
+
+struct relocate_on_75 : public relocate_on<4U> {};
+
+struct relocate_on_80 : public relocate_on<5U> {};
+
+struct relocate_on_85 : public relocate_on<7U> {};
+
+struct relocate_on_90 : public relocate_on<10U> {};
+
+template<class GrowthType, class StoredSizeType, std::size_t FreeFraction>
+struct devector_opt
+   : vector_opt<GrowthType, StoredSizeType>
+{
+   BOOST_STATIC_CONSTEXPR std::size_t free_fraction = FreeFraction;
+};
+
+typedef devector_opt<void, void, 0u> devector_null_opt;
+
+#else
+
+//!This relocation condition option specifies that the container will never relocate
+//!elements when there is no space at the side the insertion should
+//!take place
+struct relocate_never;
+
+//!This relocation condition option specifies that the container will relocate
+//!all elements when there is no space at the side the insertion should
+//!take place and memory usage is below 66% (2/3)
+struct relocate_on_66;
+
+//!This relocation condition option specifies that the container will relocate
+//!all elements when there is no space at the side the insertion should
+//!take place and memory usage is below 75% (3/4)
+struct relocate_on_75;
+
+//!This relocation condition option specifies that the container will relocate
+//!all elements when there is no space at the side the insertion should
+//!take place and memory usage is below 80% (4/5)
+struct relocate_on_80;
+
+//!This relocation condition option specifies that the container will relocate
+//!all elements when there is no space at the side the insertion should
+//!take place and memory usage is below 85% (6/7)
+struct relocate_on_85;
+
+//!This relocation condition option specifies that the container will relocate
+//!all elements when there is no space at the side the insertion should
+//!take place and memory usage is below 90% (9/10)
+struct relocate_on_90;
+
+#endif
+
+
 //! Helper metafunction to combine options into a single type to be used
 //! by \c boost::container::devector.
-//! Supported options are: \c boost::container::growth_factor and \c boost::container::stored_size
+//! Supported options are: \c boost::container::growth_factor, \c boost::container::stored_size
+//! and \c boost::container::relocate_on
 #if defined(BOOST_CONTAINER_DOXYGEN_INVOKED) || defined(BOOST_CONTAINER_VARIADIC_TEMPLATES)
 template<class ...Options>
 #else
@@ -469,7 +576,9 @@ struct devector_options
       #endif
       >::type packed_options;
    typedef devector_opt< typename packed_options::growth_factor_type
-                       , typename packed_options::stored_size_type> implementation_defined;
+                       , typename packed_options::stored_size_type
+                       , packed_options::free_fraction
+                       > implementation_defined;
    /// @endcond
    typedef implementation_defined type;
 };
@@ -497,9 +606,9 @@ using devector_options_t = typename boost::container::devector_options<Options..
 template<std::size_t BlockBytes, std::size_t BlockSize>
 struct deque_opt
 {
-   static const std::size_t block_bytes = BlockBytes;
-   static const std::size_t block_size  = BlockSize;
-   BOOST_STATIC_ASSERT_MSG(!(block_bytes && block_size), "block_bytes and block_size can't be specified at the same time");
+   BOOST_STATIC_CONSTEXPR std::size_t block_bytes = BlockBytes;
+   BOOST_STATIC_CONSTEXPR std::size_t block_size  = BlockSize;
+   BOOST_CONTAINER_STATIC_ASSERT_MSG(!(block_bytes && block_size), "block_bytes and block_size can't be specified at the same time");
 };
 
 typedef deque_opt<0u, 0u> deque_null_opt;
@@ -508,7 +617,7 @@ typedef deque_opt<0u, 0u> deque_null_opt;
 
 //! Helper metafunction to combine options into a single type to be used
 //! by \c boost::container::deque.
-//! Supported options are: \c boost::container::block_bytes
+//! Supported options are: \c boost::container::block_bytes and \c boost::container::block_size
 #if defined(BOOST_CONTAINER_DOXYGEN_INVOKED) || defined(BOOST_CONTAINER_VARIADIC_TEMPLATES)
 template<class ...Options>
 #else
